/********************************************************************************

        Program: echoart.c
        Author: Dennis Opacki <dopacki@adotout.com>
        Date: Fri May 14 10:04:27 EDT 2004
        Purpose: Respond to icmp echo requests with ASCII art

        compile with:
        gcc -o echoart -lpcap -lnet echoart.c

        command-line options:
    
        -i ifname       specify interface to listen on (default lets pcap pick)
        -t n            reset ascii drawing after 'n' seconds of inactivity
			(default 10)  
        -x n            allocate 'n' buckets
                        (default 50)
        -f              run in foreground
        -d              run in background in "daemon" mode
	-a		filename containing ascii art
        -h              usage info

    Copyright (C) 2004  Dennis Opacki
        
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
        
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License   
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

********************************************************************************/

#include <libnet.h>
#include <pcap.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h> 
#include <signal.h>
#include <syslog.h>
#include "echoart.h"

// global variables
char VERSION[] = "0.1";
struct bucket **bb;
char *asciiart = NULL;
u_int artsize = 0;
int option_t = 10;
int option_x = 50;
int option_f = 0;
int option_d = 0;
int option_h = 0;

// purge and initialize all buckets
void init_buckets() {
	u_int i;
	
	if ( ( bb = malloc( sizeof(struct bucket *) * (option_x+1)) ) == NULL ) {
		malloc_fail("bb", sizeof(struct bucket *) * (option_x+1));
	}
	for (i=0;i<=option_x;i++) {
		if ( ( bb[i] = (struct bucket *)malloc( sizeof(struct bucket) ) ) == NULL) {
			malloc_fail("bb[i]", sizeof(struct bucket) );
		}
		bb[i]->ip_addr=NULL;
		scour_bucket(i);
	}
}

// clean out a bucket
int scour_bucket(int i) {

	if (bb[i]->ip_addr != NULL) {
		free (bb[i]->ip_addr);
	}
	bb[i]->ip_addr=NULL;
	bb[i]->next=0;
	bb[i]->last=(time_t)0;

	return 1;
}

// get next response for bucket
int next_for_bucket( char * ip_src ) {
	int bucket = 0;
	int resp   = 0;

	bucket = find_bucket(ip_src);

	// set bucket fields
	if ((time(0) - bb[bucket]->last)> option_t) bb[bucket]->next=0;
	bb[bucket]->last= time(0);

	// get next character
	if ( asciiart[ ( bb[bucket]->next++) ] == '!' ) {
		resp = 1;
	}	

	// wrap if necessary
	if (bb[bucket]->next>=artsize) bb[bucket]->next=0;
	return resp;
}

// find or create bucket for this source
int find_bucket( char *ip_src) {
	int i, bucket=0;

	// look for an existing bucket for this IP
	for (i=0; i<option_x; i++) {
		if ((bb[i]->ip_addr != NULL)&&(strncmp(bb[i]->ip_addr, ip_src, strlen(bb[i]->ip_addr))==0)) {
			return i;
		}
	}

	// next look for unused buckets
	for (i=0; i<option_x; i++) {
		if (bb[i]->ip_addr == NULL) {
			scour_bucket(i);
			if ( ( bb[i]->ip_addr = (char *)strdup(ip_src) ) == NULL) malloc_fail("bb[i]->ip_addr", strlen(ip_src) );
			return i;
		}
	}

	// use one older that 10 seconds
	for (i=0; i<option_x; i++) {
		if (bb[i]->last > option_t) {
			scour_bucket(i);
			if ( ( bb[i]->ip_addr = (char *)strdup(ip_src) ) == NULL) malloc_fail("bb[i]->ip_addr", strlen(ip_src) );
			return i;
		}
	}

	// sorry, out of buckets - you get the trash pail
	bucket = option_x+1;	
	scour_bucket(bucket);
	return bucket;;
}
	
void handle_IP(u_char *args, const struct pcap_pkthdr* pkthdr,const u_char* packet){
	const struct libnet_ipv4_hdr *ip;
	const struct libnet_icmpv4_hdr *icmp;
	u_int length = pkthdr->len;
	u_int caplen = pkthdr->caplen;
	u_int hlen,off,version,len,dlen;
	char *ip_src;
	char *data;

	// skip ethernet header
	length -= sizeof(struct libnet_802_3_hdr);

	// make sure packet is valid length
	if (length < sizeof(struct libnet_ipv4_hdr)) {
		return;
	}
	
	// snap off IP portion
	ip = (struct libnet_ipv4_hdr*)(packet + sizeof(struct libnet_802_3_hdr));

        // get utility params for sanity checking
        len     = ntohs(ip->ip_len);
        hlen    = ip->ip_hl;
        version = ip->ip_v;
         
        // let's not do ipv6 just yet
        if(version != 4) {
                return;
        }
        
        // make sure we have a sane header length
        if(hlen < 5 ) {
                return;
        }

        // do we have the everything we are supposed to?
        if(length < len) {
                return;
        }
        
        // make sure we are only processing the first fragment
        off = ntohs(ip->ip_off);
        if((off & 0x1fff) == 0 ) {
 
                // get the source ip as a string (probably more efficient to use decimal)
                ip_src = (char *)inet_ntoa(ip->ip_src);

		// only interested in ICMP
		if (ip->ip_p == 1) {

			icmp = (struct libnet_icmpv4_hdr *) ( (char*) packet + sizeof(struct libnet_802_3_hdr) + sizeof (struct libnet_ipv4_hdr) );
			
			// really only want ECHO REQUEST packets
			if  ( (icmp->icmp_type == ICMP_ECHO) && (icmp->icmp_code == 0) ) {

				// echo replies must contain the data field from the original echo
				data = (u_char *) ( packet + sizeof(struct libnet_802_3_hdr) + sizeof (struct libnet_ipv4_hdr) +8);
				dlen = length- (sizeof (struct libnet_ipv4_hdr)+8);
				send_reply((char *)inet_ntoa(ip->ip_src),ntohs(icmp->icmp_id),ntohs(icmp->icmp_seq),ip->ip_tos,data,dlen);
			}
		}
	}
	return;
}

int send_reply(char *ip_dst, u_short id, u_short seq, u_char tos, char *data, u_int dlen) {
	int packet_size, c;
	libnet_ptag_t icmp,ip;
	char err_buff[LIBNET_ERRBUF_SIZE];
	int8_t *device = NULL;
	libnet_t *l;
	u_long dst_ip;

	// instantiate libnet context
	if ( ( l= libnet_init(LIBNET_RAW4, device, err_buff))==NULL) {
		printf("error creating libnet context\n");
		exit(0);
	}

	// prepare destination address
	if ((dst_ip = libnet_name2addr4(l,ip_dst, LIBNET_RESOLVE))==-1) {
		printf("Bad destination address: %s\n",ip_dst);
		exit(0);
	}

	// prepare icmp 
	if ( (icmp = libnet_build_icmpv4_echo( ICMP_ECHOREPLY,0,0,id,seq,data,dlen,l,0)) == -1) {
		printf("error creating icmp packet\n");
		exit(0);
	}

	// prepare ip packet
	if ( (ip = libnet_build_ipv4(LIBNET_IPV4_H+LIBNET_ICMPV4_ECHO_H+dlen,tos,0,0,64,IPPROTO_ICMP,0,0,dst_ip,NULL,0,l,0)) == -1) {
		printf("error creating ip packet\n");
		exit(0);
	}

	// see if we respond to this particular ping
	if ( next_for_bucket(ip_dst) ) {
//		printf("send ip_dst [%s] seq [%hu]\n",ip_dst,ntohs(seq));		
		libnet_write(l);
	}
	else {
//		printf("drop ip_dst [%s] seq [%hu]\n",ip_dst,ntohs(seq));		
	}

	// destroy libnet context
	libnet_destroy(l);

	return 1;
}

int main (int argc, char **argv) {
	char *dev  = NULL;
	char errbuf[PCAP_ERRBUF_SIZE];
	pcap_t *descr;
	struct in_addr addr;
        struct sigaction sa;
	char *args = NULL;
	char *filename = NULL;
	char *filter = "icmp[0]=8";
	struct bpf_program fp;
        bpf_u_int32 maskp=0;          /* subnet mask               */
        bpf_u_int32 netp=0;           /* ip                        */
	u_int f_size;
	int fd;
	u_int c = 0;

	// loop through command line options
	while (1) {
		int option_index = 0;
		c = getopt(argc,argv,"i:t:x:a:fdh");
		if (c==-1) break;
		switch(c) {
			case 0:
				break;
			case 'i':
				if (optarg) {
					if ( ( dev = (char *)strdup(optarg) ) == NULL) malloc_fail("dev", strlen(optarg) );
				}
				break;
			case 't':
				if (optarg) {
                                        if ( abs (atoi(optarg)) > 0) {   
                                                option_t = abs( atoi(optarg));
                                        }
				}
				break;
			case 'x':
				if (optarg) {
                                        if ( abs (atoi(optarg)) > 0) {   
                                                option_x = abs( atoi(optarg));
                                        }
				}
				break;
			case 'a':
				if (optarg) {
					if ( ( filename = (char *)strdup(optarg) ) == NULL) malloc_fail("filename", strlen(optarg) );
				}
				break;
                        case 'f':
                                option_f = 1;
                                break;
                        case 'd':
                                option_d = 1;
                                break;
                        case 'h':
                                option_h = 1;
                        default:
                                break;
		}
	}

        // display usage info if needed
        if (optind<argc) option_h = 1;
        if (option_h) {
                fprintf(stderr,"echoart, version %s\n",VERSION);
                fprintf(stderr,"Usage: %s [OPTION]\n\n",argv[0]);
                fprintf(stderr,"-i IFNAME               specify device name to listen on\n");
                fprintf(stderr,"-t N                    reset ascii drawing after N seconds idle\n");
                fprintf(stderr,"-x N                    create N buckets\n");
                fprintf(stderr,"-a FILENAME             name of file containing ascii art\n");
                fprintf(stderr,"-f                      run in foreground mode\n");
                fprintf(stderr,"-d                      run in daemon mode\n");
                fprintf(stderr,"-h                      display this usage information\n");
                exit(1);
        }                        
        
        if ( ( ! option_d ) && ( ! option_f ) ) {
                fprintf(stderr,"%s couldn't start\n",argv[0]);
                fprintf(stderr,"You must specify either either -d (daemon) or -f (foreground)\n");
                exit(1);
        }

	// open input file and read data
	if ((fd = open(filename, O_RDONLY)) == -1){
		fprintf(stderr,"unable to open file\n");
		exit(1);
	} 
	else {
		asciiart = full_read(fd,&artsize);
		close(fd);
	}

	// prepare bucket brigade
	init_buckets();

        // find a valid device to open
        if(dev == NULL && ( (dev=pcap_lookupdev(errbuf)) == NULL ) ){
                fprintf(stderr,"unable to bind to valid device\n");
                exit(1);
        }               

        // get network address and netmask for device
        pcap_lookupnet(dev,&netp,&maskp,errbuf);
 
        // open device for reading only local traffic
        descr = pcap_open_live(dev,1500,0,1,errbuf);   
        if(descr == NULL) {
                fprintf(stderr,"unable to open device %s\n",dev);
                exit(1);
        }

        // compile filter
        if(pcap_compile(descr,&fp,filter,0,netp) == -1) {
                fprintf(stderr,"unable to compile filter[%s]\n",filter);
                exit(1);
        }
        
        // set filter
        if(pcap_setfilter(descr,&fp) == -1){
		fprintf(stderr,"unable to apply filter\n");
                exit(1);
        }
 
        // set up for daemonized operation unless running in foreground
        if ( ! option_f ) {
                openlog("echoart",LOG_PID|LOG_CONS,LOG_DAEMON);
                syslog(LOG_NOTICE,"echoart starting");

                // daemonize unless running in foreground
                daemonize();

                // set up signal handlers  
                sa.sa_handler=exit;
                sa.sa_flags=0;
                if(sigaction(SIGTERM,&sa,NULL)) {
                        syslog(LOG_ERR,"Unable to set signal handler: %s.  Exiting.",
                        strerror(errno));
                }
        }
  
        // main pcap loop   
        pcap_loop(descr,-1,handle_IP,args);
}

// linked list structure for full_read
static struct buff_head *newbuf(void)
{
  struct buff_head *new;
                                
  new=malloc(sizeof(struct buff_head));
  new->buf=malloc(BUFFERSIZE);   
  new->len=0;
                                
  return new;
}

// read an as-yet undetermined amount of data from a file descriptor
char *full_read (int fd, int *size) {
        int total = 0;
        int thistime;
        int offset;
        struct buff_head *head,*buf;
        char *result;
                
        head=newbuf();
        buf=head;
        
        for (;;) {
                thistime = read (fd, buf->buf, BUFFERSIZE);
                buf->len=thistime;
                
                if (thistime < 0) {
                        size=0;
                        return NULL;
                }   
                else if (thistime == 0) {
                        /* EOF, but we didn't read the minimum.  return what we've read
                         * so far and next read (if there is one) will return 0. */
                        break;
                }
                
                total += thistime;
                buf->next=newbuf();
                buf=buf->next;
        }
        
        result=malloc(total+1);
        offset=0;

        for(buf=head;buf;buf=buf->next) {
                memcpy(result+offset,buf->buf,buf->len);
                offset+=buf->len;
        }

        *size=total;

        return result;
}

int malloc_fail( char * var, int size ) {
        // print error to stderr if running in bindsnap mode
        if (option_f) {
                fprintf(stderr, "our OS wouldn't let me malloc %d bytes for a new %s. giving up", size, var);
        }
        else {
                syslog(LOG_ERR, "our OS wouldn't let me malloc %d bytes for a new %s. giving up", size, var);
        }       
        exit(1);
}       

// daemonize the process
int daemonize(void) {
        pid_t pid;
        int fd;

        fd=open("/dev/null",O_RDWR);
        if(fd<0) {
                syslog(LOG_ERR,"Failed to open /dev/null: %s.  Exiting.",strerror(errno));
                exit(1);
        }

        dup2(fd,0);
        dup2(fd,1);
        dup2(fd,2);

        if((pid=fork())<0) {
                syslog(LOG_ERR,"Fork failed: %s.  Exiting.",strerror(errno));
                exit(1);
        }
        else if (pid!=0) {
                exit(0);
        }

        setsid();
        chdir("/");
        umask(0);
        return 0;
}

